/* nu_41    "Production version"
    Hamster Wheel Tachometer project
    Code name: nugget

   Adafruit p/n 2771 Feather 32u4
   Adafruit p/n 3129 Alphanumeric LED feather wing
   Vishay   p/n TSSP4038 IR remote receiver IC sensor
   Vishay   p/n TSAL4400 IR LED

  Adafruit #2771 32u4 Feather Pinout
             +--------| USB |--------+
             |RST                    |
  display    |3V3                    |
             |AREF                   |
  display    |GND                    |
             |A0                 VBAT|
             |A1                   EN|
             |A2                 VBUS|
             |A3                  D13|  TSSP_DATA (!PCBA_LED)  Must be D13!!
             |A4                  D12|
             |A5                  D11|  TSSP_VCC
             |D15 SCLK            D10|  IR_LED   Must be D10!!
             |D16 MOSI             D9|  debug sygnal
             |D14 MISO             D6|
             |D0 RXD INT2          D5|
     SW_PB   |D1 TXD INT3 INT0 SCL D3|  display
             |GND         INT1 SDA D2|  display
             +-----------------------+
*/
#include <avr/wdt.h>
#include <LowPower.h>
#include <EEPROM.h>
#define SW_PB 0         // push button switch 
enum mode { COUNT, CAL, DISPLY } nuggetMode;  // three states of operation
bool TRUE = 1;          //
bool FALSE = 0;         //
#define sygnal 9        // pin used to show loop operation (heart beat)

/**************************** Remember ***************************************
   Static variables are visible to only one function, but preserves data
   between function calls. Variables declared as static will only be created
   and initialized the first time a function is called.
   Volatile variables are in RAM (not a storage register). Used both within an
   ISR and outside it.
*/
// display stuff --------------------------------------------------------------
#include "Wire.h"
#include <Ada14seg_table.h>
const uint8_t displayAddr = 0x70;       // HT16K33 default display address
String buff = "";                       // allocate display buffer
const uint8_t displayBrightness = 12;   // range 0-15
uint16_t displayBuffer[4];              // four digit display (no scrolling)

// timer 1 variables (38kHz bursts)--------------------------------------------
#define IR_LED 10            // !!! OC1B  must use this pin 
#define TSSP_VCC 11          // power pin for tssp4038
#define burstCount 25        // actual number of 38kHz pulses
#define burstAddon 20        // add for sensor's end of burst response time
int8_t volatile static  burstCounter = 0;   // counts timer 1 38kHz cycles

// timer 3 variables (TSSP pulse width capture)-------------------------------
#define TSSP_DATA 13                       // !!! ICP3 must use this pin!!
volatile static unsigned int icrTemp;      // captures ICR before it changes in timer3 isr
volatile static unsigned int startingEdge; // timer3 count at beginning of TSSP edge
volatile static unsigned int endingEdge;   // timer3 count at end of TSSP edge
volatile static unsigned int pulseWidth;   // length of TSSP pulse

// hamster wheel revolution variables -----------------------------------------
#define INTERVALS 1547 // from MEASUREMENTS on prototypes using sygnal pin
/* A complication we will ignore:
   The range setting changes the timer1 freq, which changes the burst time,
   which changes the loop time and thus affects the actual one minute time.
   The WDT is good for +/-10% anyway so who cares.
*/
uint16_t static wheelCounter = 0;             // number of wheel revs
uint16_t static previousPw = 0;               // TSSP value last wheel check
uint16_t volatile nugCounterMax = 0;          // max number of revs in 1 min
uint16_t volatile nugCounterTotal = 0;        // total revs since last reset
uint16_t volatile intervalCounter = INTERVALS;// keeps track of how many wdts

// switch variables related to calibration ------------------------------------
uint16_t volatile static  wdtTicCounter = 0;
uint8_t swState = 0;        // starting with state 0: waiting for button press
unsigned long startTime;    // when pushbutton down will be set in state 1
uint16_t pressTics = 20;    // number of wdt for a pb press 30mS * 20 = 600 mS
uint16_t debounceTics = 1;  // changes less than one wdt are ignored.

// calibration data -----------------------------------------------------------
//
uint8_t EEPROMAddr = 0; // where in eeprom to store the user selected cal range
struct CalData {        // each calibration range needs 3 pieces of data
  uint8_t ocr1a, ocr1b; // the timer1 settings for freq and duty cycle
  char symbol;          // and a symbol to display so we know which range
};

CalData cal_ranges[3] = {
  {125, 15, 'L'},       // LR = 1.5 to 5"   32 kHz  12% Duty Cycle
  {135, 15, 'M'},       // MR = 1.3 to 3.5" 29.6kHz 11% Duty Cycle
  {145, 10, 'S'}        // SR = 0.5 to 2"   27.6kHz  7% Duty Cycle
};
uint8_t calDataIndex = 0;

// ============================================================================
void setup() {
  wdt_disable();                // makes uploads easy to accomplish
  delayMicroseconds(5000);      // a little time w/o using timers
  //------------------------------------------------------------------
  pinMode(IR_LED, OUTPUT);      // controlled through timer 1 OCR1B
  digitalWrite(IR_LED, LOW);    //
  pinMode(TSSP_VCC, OUTPUT);    // switched on in wheelCheck
  digitalWrite(TSSP_VCC, LOW);  // off for now
  pinMode(TSSP_DATA, INPUT);    // must use this hardware pin for ICP3
  pinMode(SW_PB, INPUT_PULLUP); // can really be assigned most anywhere
  pinMode(sygnal, OUTPUT);      // generic debug pin toggles in loop{}
  digitalWrite(sygnal, HIGH);   //  interrupts();                 //
  //---------------------------------------------------------------------------
  // display init and state
  displayOn();
  displaySetBrightness(displayBrightness);
  if (digitalRead(SW_PB) == 0) {    // pb down at startup -> calibrate mode
    nuggetMode = CAL;
    displayString("Cal", SLEEP_2S);
  } else {                          // else normal operation
    nuggetMode = COUNT;
    displayString("nu41", SLEEP_2S);
  }
  displayClear();
  /*---------------------------------------------------------------------------
    init EEPROM and counters
    displaydisplayAddr 0 cal values 0=Long range, 1=Medium, 2=short range
  */
  if (EEPROM.read(EEPROMAddr) == 0xff) {   // cal data never been set
    EEPROM.update(EEPROMAddr, 0);          // so make range=Long
  }
  calDataIndex = EEPROM.read(EEPROMAddr);
  OCR1A = cal_ranges[calDataIndex].ocr1a;     // load previous range values
  OCR1B = cal_ranges[calDataIndex].ocr1b;     //
  displayString("Ra=" + String(cal_ranges[calDataIndex].symbol), SLEEP_2S);
}

// ============================================================================
void loop() {
  /* Three states or modes of operation: COUNT,CAL,DISPLY.
     COUNT - normal operation counting the exercise wheel revs
        Enter on power up or after DISPLY.
        Exit for push button (PB) for a DISPLY.
        Action - checks if wheel moved, if so updates now and max counts.
              checks if one minute has passed, if so updates total count.
              activates watch dog timer power down ... sleepy time
              checks if PB down, if so changes mode to DISPLY

     CAL - User selects best range given distance between exercise wheel and HWT
        Enter on power up if PB down.
        Exit on power off
        Action - little state machine based on swState and push button (PB)
          swState pbUpDown   Action
            -------------------------------------------------------
            0        0    user has PB down start counter to see for how long
            0        1    fall through to wheel check.
            1        0    PB still down
            1        1    how long was PB down?
                                1 tic = bouncy PB so swState back to 0
                                < x tics = next cal range
                                > x tics = user wants to save this range
                          swState back to 0
          Wheel check - measure pulse width. WDT for 64mS

     DISPLY - guess what? displays the data (now revs, max RPM, total revs)
        Enter on mode DISPLY
        Exit when done with displaying.
        Action - displays
                 change mode to COUNT
  */

  digitalWrite(sygnal, !digitalRead(sygnal));  // use to measure freq of loop
  uint8_t pbUpDown;                            //
  switch (nuggetMode) {                        //
    case COUNT:  //------------------------------------------------------------
      WheelCheck();                            //
      if (previousPw > 18 && (pulseWidth == 0)) {  // light/dark happened
        wheelCounter++;                            // wheel turned one rev.
      }
      if (intervalCounter == 0) {          // one min elapsed so update counters
        displayString("_   ", SLEEP_15MS); // heartbeat every min = hwt working
        if (wheelCounter > nugCounterMax) {
          nugCounterMax = wheelCounter;
        }
        nugCounterTotal = nugCounterTotal + wheelCounter;
        wheelCounter = 0;              // reset values
        intervalCounter = INTERVALS;   // counts wdts in one minute
      } else  {                        // not one min yet
        displayClear();                // make sure heartbeat off
        intervalCounter--;             // one less wdt until a minute
      }
      previousPw = pulseWidth; // save pw so we can tell if light/dark transition occurred
      LowPower.powerDown(SLEEP_30MS, ADC_OFF, BOD_OFF);     // sleepy time
      if (digitalRead(SW_PB) == 0) {   // user wants to display data
        nuggetMode = DISPLY;           // new state when next we do switch-case
      }
      break;

    case CAL:
      wdtTicCounter++;      // counts number of wdts (typically 64mS)
      // state machine for push button  ---------------------------------------
      pbUpDown = digitalRead(SW_PB); // current push button signal
      if (swState == 0) {            // default = just a wheel check and display
        if (pbUpDown == 0) {         // Oh oh, user signals an action needed
          swState = 1;               // step to state 1
          startTime = wdtTicCounter; // remember starting time
        }
      } else if (swState == 1) {     // user has pressed the PB previously
        // PB was released too quickly need debouncing. back to state 0
        if ((pbUpDown) && ((unsigned long)(wdtTicCounter - startTime) < debounceTics)) {
          swState = 0;
          // PB is up so user wants the next range (L,M,S)
        } else if (pbUpDown == 1) {
          if (calDataIndex < 2) {
            calDataIndex++;
          } else {
            calDataIndex = 0;
          }
          OCR1A = cal_ranges[calDataIndex].ocr1a;  //  get new set of cal data
          OCR1B = cal_ranges[calDataIndex].ocr1b;  //
          //wdtTicCounter = 0;          // reset the tic counter
          buff = String (cal_ranges[calDataIndex].symbol) + "   ";
          wdtTicCounter = 0;      // done with range change - reset
          swState = 0;            // reset
        }
        //  PB has been down for a while- user wants to save this range setting
        else if ((pbUpDown == 0) && (wdtTicCounter > startTime + pressTics)) {
          EEPROM.update(EEPROMAddr, calDataIndex); // save that cal selection
          displayString(String("Savd"), SLEEP_2S);  // tell the user
          wdtTicCounter = 0;      // done with save - reset
          swState = 0;            // reset
        }
      }   // end of state machine   -------------------------------------------
      // always do a pulse width check in the CAL loop
      WheelCheck();
      if (pulseWidth == 0) {   // display the readings from the TSSP sensor
        buff = String (cal_ranges[calDataIndex].symbol) + "   ";
      } else {
        buff = String (cal_ranges[calDataIndex].symbol) + ' ' + String(pulseWidth);
      }
      displayString(buff, SLEEP_60MS); // wait longer in the CAL loop than in a
      // COUNT loop. Reduces display flicker.
      break;

    case DISPLY: //------------------------------------------------------------
      displayString("Now=", SLEEP_1S);
      displayClear();
      displayString(String(wheelCounter), SLEEP_1S);
      displayString("Max=", SLEEP_1S);
      displayClear();
      if (wheelCounter >= nugCounterMax) {
        displayString(String(wheelCounter), SLEEP_1S);
      } else {
        displayString(String(nugCounterMax), SLEEP_1S);
      }
      displayString("Tot=", SLEEP_1S); displayClear();
      displayString(String(nugCounterTotal), SLEEP_1S);
      displayClear();
      nuggetMode = COUNT;
      break;

    default: //----------------------------------------------------------------
      break;
  }
}
/*================== wheel check function =====================================
  Do one IR burst from LED and monitor TSSP sensor output.

    Timer1 generates the ≈38kHz pulses:
      Timer mode 9 PWM phase and freq correct
      From the calibration structure:
        OCR1A = sets frequency
        OCR1B = sets duty cycle

    Timer 1 overflow ISR counts the number of pulses:
        There are xx pulses then Timer 1 output is disconnected. xx=burstCount
        Then there are yy more pulses counted to give time for the TSSP sensor
        to respond to the ≈38kHz burst.  yy=burstAddon

    Timer3 counts the TSSP output pulse width (active low)
        Timer mode 0 normal. Input clock = clk/256 (happens to be about 31kHz)

    Timer3 capture ISR is used to measure the pulse width.
        On the falling edge the ISR captures the timer3 count then sets
            timer3 to capture next on rising edge.
        On the rising edge ISR captures the ending count.

*/
void WheelCheck (void) {
  startingEdge = 0;       	//
  endingEdge = 0;         	//
  pulseWidth = 0;         	//
  icrTemp = 0;            	// temp storage of the input capture register value
  // --------------------------------------------------------------------------
  digitalWrite(TSSP_VCC, HIGH);  // turn on TSSP sensor
  delayMicroseconds(2500);       // delay for TSSP stability-doesn't use timers!
  //---------------------------------------------------------------------------
  // Timer 1 burst reset
  burstCounter = burstCount  + burstAddon;    //
  cli();                        // no interrupts for the moment
  TCCR1A = B00100001;           // TCCR1A = 0x21  frequency
  TCCR1B = B00010001;           // TCCR1B = 0x11  duty cycle
  TIMSK1 |= (1 << TOIE1);       // enable timer 1 overflow isr

  /* Unless the register has read only bits, I prefer to write register bytes.
    I have the datasheet out anyway to keep track of all the bit acronyms. Bad
    habit I picked up in the last century.
  */
  //---------------------------------------------------------------------------
  // Timer 3 sensor pw reset
  cli();
  TCCR3A = 0x00;                //
  TCCR3B = 0x84;                // input capture on falling edge in input pin
  TIMSK3 |= (1 << ICIE3);       // enable input capture isr
  sei();                        // interrupt away
  //---------------------------------------------------------------------------
  //  stay here while the burst happens
  while (burstCounter != 0) {
  }                             // hmm, what if we get stuck? kaboom
  TCCR3B = 0x00;                // turn off timer 3 (no clock source)
  TCCR1B = 0x00;                // turn off timer 1 (no clock source)
  TIMSK3 &= ~(1 << TOIE3);      // disable overflow isr
  //TIMSK3 &= ~(1 << ICIE3);    // disable input capture isr
  //----------------------------------
  // As long as there was a falling (startingEdge) and rising (endingEdge) on
  // the TSSP output calculate the sensor pulse width.
  pulseWidth = endingEdge - startingEdge;
  digitalWrite(TSSP_VCC, LOW);   // TSSP turned off now
}

ISR(TIMER1_OVF_vect) {  //****************************************************
  // comes here with every ≈38kHz pulse. Counts pulses and controls
  // timer1 output
  burstCounter --;
  if (burstCounter <= burstAddon) {
    TCCR1A &= ~(1 << COM1B1 ) ;  // disconnect timer output pin
  }
  if (burstCounter <= 0) {
    TCCR1B &= ~(1 << CS10 ) ;    // turn off timer
  }
}

ISR(TIMER3_CAPT_vect) {  //****************************************************
  // timer 3 captures the TSSP sensor output pulse times
  icrTemp = ICR3;             // quick --> capture counter
  cli();                      // hold off
  if ( TCCR3B == 0xC4) {      // here because of rising edge of pw
    endingEdge = icrTemp;     // get end time for period
  }
  if (TCCR3B == 0x84) {       // here because falling edge of pw
    startingEdge = icrTemp;   // get beginning time for period
    TCCR3B = 0xC4;            // now we look for rising edge
  }
  sei();                      // interrupt away
}

/*====================== display functions ====================================
  Based upon:
  www.partsnotincluded.com/electronics/controlling-led-matrix-with-the-ht16k33
  https://github.com/dmadison/LED-Segment-ASCII
  https://github.com/lpaseen/ht16k33

  Another bad habit... I take perfectly good libraries and savage them.
  All because I want some little change.

  Once a string is written to the display, the 32U4 uses the WDT to power
  down for power savings. While a display is happening the exercise wheel is
  ignored. The loop {} minute timer does NOT account for these seconds spent
  displaying. I know you'll do better.
*/

void displayShow() { // -------------------------------------------------------
  Wire.beginTransmission(displayAddr);
  Wire.write(0x00);                       // start at displayAddress 0x0
  for (int i = 0; i < 4; i++) {
    Wire.write(displayBuffer[i] & 0xFF);
    Wire.write(displayBuffer[i] >> 8);
  }
  Wire.endTransmission();
}

void displayClear() { // ------------------------------------------------------
  for (int i = 0; i < 4; i++) {
    displayBuffer[i] = 0;
  }
  displayShow();
}

void displaySetBrightness(uint8_t b) {  // ------------------------------------
  if (b > 15) return;
  Wire.beginTransmission(displayAddr);
  Wire.write(0xE0 | b);                 // Dimming command
  Wire.endTransmission();
}

void displayWriteChar(uint8_t dispNum, char c) { // ---------------------------
  displayBuffer[dispNum] = pgm_read_word_near(Ada14seg_table + (c - 32));
}

void displayString(String buff, period_t sleep_time) { // ---------------------
  displayOn();
  int dispBuffLength = buff.length();
  for (int i = 0; i <= dispBuffLength; i++) {
    displayWriteChar(4 - dispBuffLength + i, buff[i]);
  }
  displayShow();
  LowPower.powerDown(sleep_time, ADC_OFF, BOD_OFF); // Sleep and display
  displayOff();
}

void displayOn() { // ---------------------------------------------------------
  Wire.beginTransmission(displayAddr);
  Wire.write(0x21);               // turn on oscillator
  Wire.endTransmission();

  Wire.beginTransmission(displayAddr);
  Wire.write(0x81);               // no blinking / blanking command
  Wire.endTransmission();
}

void displayOff() { // --------------------------------------------------------
  Wire.beginTransmission(displayAddr);
  Wire.write(0x20);               // turn off oscillator
  Wire.endTransmission();
}
